/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.awt.Color;

/**
   Rasterize a {@link LineSegment} into pixels in the
   {@link FrameBuffer}'s viewport and anti-alias the
   line at the same time.
<p>
   This is a very simplistic anti-aliasing algorithm.
   It only works with white lines on a black background.
*/
public class RasterizeAntialias
{
   public static boolean debug = false;

   /**
      Rasterize a {@link LineSegment} into pixels in the
      {@link FrameBuffer}'s viewport.

      @param ls  {@link LineSegment} to rasterize into the {@link FrameBuffer}
      @param fb  {@link FrameBuffer} to hold rasterized pixels
   */
   public static void rasterize(LineSegment ls, FrameBuffer fb)
   {
      // Make local copies of several values.
      int w = fb.getWidthVP();
      int h = fb.getHeightVP();

      Vertex v0 = ls.model.vertexList.get( ls.index[0] );
      Vertex v1 = ls.model.vertexList.get( ls.index[1] );

      // Round the line segment's two endpoints to the nearest
      // logical pixel. This makes the algorithm a lot simpler,
      // but it can cause a slight, but noticable, shift of the
      // line segment.
      double x0 = Math.round(v0.x),  y0 = Math.round(v0.y);
      double x1 = Math.round(v1.x),  y1 = Math.round(v1.y);

      // Rasterize a degenerate line segment (a line segment
      // that projected onto a point) as a single pixel.
      if ( (x0 == x1) && (y0 == y1) )
      {
         // We don't know which endpoint of the line segment
         // is in front, so just pick v0.
         fb.setPixelVP((int)x0 - 1, h - (int)y0, Color.white);
         return;
      }

      if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
      {
         if (x1 < x0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute this line segment's slope.
         double m = (y1 - y0) / (x1 - x0);

         // Rasterize this line segment.
         // In the following loop, as x moves across the logical
         // horizontal pixels, we will compute a y value for each x.
         double y = y0;
         for (int x = (int)x0; x < (int)x1; x++, y += m)
         {
            // Check if y is on the bottom or top edge of the pixel coordinate space.
            if ( y <= 1 || h <= y )
            {
               // Set this pixel in the framebuffer.
               fb.setPixelVP(x - 1, h - (int)Math.round(y), Color.white);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, 1.0f);
            }
            else // y must be between two logical vertical pixel coordinates
            {
               // Let y_low and y_hi be the logical vertical pixel
               // coordinates that bracket around y.
               int y_low = (int)y;                // the integer part of y
               int y_hi  = y_low + 1;
               // Let weight be the fractional part of y. We will use
               // weight to determine how much emphasis to place on
               // each of the two pixels that bracket y.
               float weight = (float)(y - y_low);
               // The smaller weight is, the closer y is to the lower
               // pixel, so we give the lower pixel more emphasis when
               // weight is small.
               float g_low = (1-weight); // g for gray scale
               float g_hi = weight;

               // Set this (antialiased) pixel in the framebuffer.
               fb.setPixelVP(x - 1, h - y_low, new Color(g_low, g_low, g_low));
               fb.setPixelVP(x - 1, h - y_hi,  new Color(g_hi,  g_hi,  g_hi));

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, g_low, g_hi);
            }
            // Advance (x,y) to the next pixel (delta_x is 1, so delta_y is m).
         }
         // Set the pixel for the (x1,y1) endpoint (we do this separately to avoid roundoff errors).
         fb.setPixelVP((int)x1 - 1, h - (int)y1, Color.white);
      }
      else  // abs(slope) > 1
      {
         if (y1 < y0) // swap (x0, y0) with (x1, y1)
         {
            double tempX = x0;
            double tempY = y0;
            x0 = x1;
            y0 = y1;
            x1 = tempX;
            y1 = tempY;
         }

         // Compute m and b in x = m*y + b.
         double m = (x1 - x0) / (y1 - y0);

         // Rasterize this line segment.
         // In the following loop, as x moves across the logical
         // horizontal pixels, we will compute a y value for each x.
         double x = x0;
         for (int y = (int)y0; y < (int)y1; x += m, y++)
         {
            // Check if x is on the left or right edge of the pixel coordinate space.
            if ( x <= 1 || w <= x )
            {
               // Set this pixel in the framebuffer.
               fb.setPixelVP((int)Math.round(x) - 1, h - y, Color.white);

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, 1.0f);
            }
            else // x must be between two logical horizontal pixel coordinates
            {
               // x_low and x_hi will be pixel coordinates that bracket around x
               int x_low = (int)x;                // the integer part of x
               int x_hi  = x_low + 1;

               float weight = (float)(x - x_low); // the fractional part of x

               float g_low = (1-weight);
               float g_hi = weight;

               // set this (antialiased) pixel in the framebuffer
               fb.setPixelVP(x_low - 1, h - y, new Color(g_low, g_low, g_low));
               fb.setPixelVP( x_hi - 1, h - y, new Color(g_hi,  g_hi,  g_hi));

               // log interesting information to standard output
               if (debug) logPixel(w, h, x, y, g_low, g_hi);
            }
            // Advance (x,y) to the next pixel (delta_y is 1, so delta_x is m).
         }
         // Set the pixel for the (x1,y1) endpoint (we do this separately to avoid roundoff errors).
         fb.setPixelVP((int)x1 - 1, h - (int)y1, Color.white);
      }
   }


   private static void logPixel(int w, int h, int x, double y, float g)
   {
      System.out.printf(
         "[w=%d, h=%d]  x=%d, y=%.3f, gray=%f\n", w, h, x, y, g);
   }

   private static void logPixel(int w, int h, double x, int y, float g)
   {
      System.out.printf(
         "[w=%d, h=%d]  x=%.3f, y=%d, gray=%f\n", w, h, x, y, g);
   }

   private static void logPixel(int w, int h, int x, double y, float g1, float g2)
   {
      System.out.printf(
         "[w=%d, h=%d]  x=%d, y=%.3f, gray_low=%f, gray_hi=%f\n", w, h, x, y, g1, g2);
   }

   private static void logPixel(int w, int h, double x, int y, float g1, float g2)
   {
      System.out.printf(
         "[w=%d, h=%d]  x=%.3f, y=%d, gray_low=%f, gray_hi=%f\n", w, h, x, y, g1, g2);
   }
}
